@@ -207,17 +207,36 @@ struct size_val
   uintmax_t size;
 };
 
+#define NEW_EXEC 1
+/*
+#undef NEW_EXEC 
+*/
+
+#if !defined(NEW_EXEC)
 struct path_arg
 {
   short offset;			/* Offset in `vec' of this arg. */
   short count;			/* Number of path replacements in this arg. */
   char *origarg;		/* Arg with "{}" intact. */
 };
+#endif
+
+#include "buildcmd.h"
 
 struct exec_val
 {
+#if defined(NEW_EXEC)
+  /* new-style */
+  boolean multiple;		/* -exec {} \+ denotes multiple argument. */
+  struct buildcmd_control ctl;
+  struct buildcmd_state   state;
+  char **replace_vec;		/* Command arguments (for ";" style) */
+  int num_args;
+  boolean use_current_dir;      /* If nonzero, don't chdir to start dir */
+#else
   struct path_arg *paths;	/* Array of args with path replacements. */
   char **vec;			/* Array of args to pass to program. */
+#endif
 };
 
 /* The format string for a -printf or -fprintf is chopped into one or
@@ -347,6 +366,7 @@ enum SymlinkOption
 extern enum SymlinkOption symlink_handling; /* defined in find.c. */
 
 void set_follow_state PARAMS((enum SymlinkOption opt));
+void cleanup(void);
 
 /* fstype.c */
 char *filesystem_type PARAMS((const struct stat *statp));
@@ -372,6 +392,7 @@ boolean pred_ctime PARAMS((char *pathname, struct stat *stat_buf, struct predica
 boolean pred_delete PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_empty PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_exec PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_execdir PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_false PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_fls PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_fprint PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
@@ -395,6 +416,7 @@ boolean pred_newer PARAMS((char *pathname, struct stat *stat_buf, struct predica
 boolean pred_nogroup PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_nouser PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_ok PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+boolean pred_okdir PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_open PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_or PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_path PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
@@ -412,7 +434,15 @@ boolean pred_uid PARAMS((char *pathname, struct stat *stat_buf, struct predicate
 boolean pred_used PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_user PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
 boolean pred_xtype PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr));
+
+int launch PARAMS((const struct buildcmd_control *ctl,
+		   struct buildcmd_state *buildstate));
+
+
 char *find_pred_name PARAMS((PFB pred_func));
+
+
+
 #ifdef DEBUG
 void print_tree PARAMS((FILE*, struct predicate *node, int indent));
 void print_list PARAMS((FILE*, struct predicate *node));
