@@ -17,6 +17,7 @@
    USA.
 */
 
+
 #include "defs.h"
 #include <ctype.h>
 #include <pwd.h>
@@ -26,6 +27,8 @@
 #include "modetype.h"
 #include "../gnulib/lib/xstrtol.h"
 #include "../gnulib/lib/xalloc.h"
+#include "buildcmd.h"
+#include "nextelem.h"
 
 
 #if ENABLE_NLS
@@ -73,6 +76,7 @@ static boolean parse_d PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_depth PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_empty PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_exec PARAMS((char *argv[], int *arg_ptr));
+static boolean parse_execdir PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_false PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_fls PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_fprintf PARAMS((char *argv[], int *arg_ptr));
@@ -104,6 +108,7 @@ static boolean parse_nogroup PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_nouser PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_nowarn PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_ok PARAMS((char *argv[], int *arg_ptr));
+static boolean parse_okdir PARAMS((char *argv[], int *arg_ptr));
 boolean parse_open PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_or PARAMS((char *argv[], int *arg_ptr));
 static boolean parse_path PARAMS((char *argv[], int *arg_ptr));
@@ -134,7 +139,7 @@ static boolean insert_regex PARAMS((char *argv[], int *arg_ptr, boolean ignore_c
 static boolean insert_type PARAMS((char *argv[], int *arg_ptr, boolean (*which_pred )()));
 static boolean insert_fprintf PARAMS((FILE *fp, boolean (*func )(), char *argv[], int *arg_ptr));
 static struct segment **make_segment PARAMS((struct segment **segment, char *format, int len, int kind));
-static boolean insert_exec_ok PARAMS((boolean (*func )(), char *argv[], int *arg_ptr));
+static boolean insert_exec_ok PARAMS((const char *action, boolean (*func )(), char *argv[], int *arg_ptr));
 static boolean get_num_days PARAMS((char *str, uintmax_t *num_days, enum comparison_type *comp_type));
 static boolean insert_time PARAMS((char *argv[], int *arg_ptr, PFB pred));
 static boolean get_num PARAMS((char *str, uintmax_t *num, enum comparison_type *comp_type));
@@ -194,6 +199,7 @@ static struct parser_table const parse_table[] =
   {ARG_OPTION,             "depth",                 parse_depth},
   {ARG_TEST,               "empty",                 parse_empty},	/* GNU */
   {ARG_ACTION,             "exec",                  parse_exec},
+  {ARG_ACTION,             "execdir",               parse_execdir},     /* *BSD, GNU */
   {ARG_TEST,               "false",                 parse_false},	/* GNU */
   {ARG_ACTION,             "fls",                   parse_fls},		/* GNU */
   {ARG_POSITIONAL_OPTION,  "follow",                parse_follow},	/* GNU, Unix */
@@ -233,6 +239,7 @@ static struct parser_table const parse_table[] =
   {ARG_PUNCTUATION,        "o",                     parse_or},
   {ARG_PUNCTUATION,        "or",                    parse_or},		 /* GNU */
   {ARG_ACTION,             "ok",                    parse_ok},
+  {ARG_ACTION,             "okdir",                 parse_okdir},        /* GNU (-execdir is BSD) */
   {ARG_TEST,               "path",                  parse_path},	 /* GNU, HP-UX, GNU prefers wholename */
   {ARG_TEST,               "perm",                  parse_perm},
   {ARG_ACTION,             "print",                 parse_print},
@@ -560,7 +567,13 @@ parse_empty (char **argv, int *arg_ptr)
 static boolean
 parse_exec (char **argv, int *arg_ptr)
 {
-  return (insert_exec_ok (pred_exec, argv, arg_ptr));
+  return (insert_exec_ok ("-exec", pred_exec, argv, arg_ptr));
+}
+
+static boolean
+parse_execdir (char **argv, int *arg_ptr)
+{
+  return (insert_exec_ok ("-execdir", pred_execdir, argv, arg_ptr));
 }
 
 static boolean
@@ -1098,7 +1111,13 @@ parse_nowarn (char **argv, int *arg_ptr)
 static boolean
 parse_ok (char **argv, int *arg_ptr)
 {
-  return (insert_exec_ok (pred_ok, argv, arg_ptr));
+  return (insert_exec_ok ("-ok", pred_ok, argv, arg_ptr));
+}
+
+static boolean
+parse_okdir (char **argv, int *arg_ptr)
+{
+  return (insert_exec_ok ("-okdir", pred_okdir, argv, arg_ptr));
 }
 
 boolean
@@ -1797,6 +1816,7 @@ make_segment (struct segment **segment, char *format, int len, int kind)
     case 'n':			/* number of links */
       fprintf_stat_needed = true;
       *fmt++ = 's';
+      break;
       
       /* Numeric items that DO honour #, 0, + flags.
        */
@@ -1813,10 +1833,195 @@ make_segment (struct segment **segment, char *format, int len, int kind)
 
   return (&(*segment)->next);
 }
+
+static void 
+check_path_safety(const char *action)
+{
+  const char *path = getenv("PATH");
+  char *s;
+  s = next_element(path, 1);
+  while ((s = next_element ((char *) NULL, 1)) != NULL)
+    {
+      if (0 == strcmp(s, "."))
+	{
+	  error(1, 0, _("The current directory is included in the PATH environment variable, which is insecure in combination with the %s action of find.  Please remove the current directory from your $PATH (that is, remove \".\" or leading or trailing colons)"),
+		action);
+	}
+    }
+}
 
+
+/* handles both exec and ok predicate */
+#if defined(NEW_EXEC)
 /* handles both exec and ok predicate */
 static boolean
-insert_exec_ok (boolean (*func) (/* ??? */), char **argv, int *arg_ptr)
+new_insert_exec_ok (const char *action,
+		    boolean (*func) (/* ??? */),
+		    char **argv, int *arg_ptr)
+{
+  int start, end;		/* Indexes in ARGV of start & end of cmd. */
+  int i;			/* Index into cmd args */
+  int saw_braces;		/* True if previous arg was '{}'. */
+  boolean allow_plus;		/* True if + is a valid terminator */
+  int brace_count;		/* Number of instances of {}. */
+  const char *prefix;
+  size_t pfxlen;
+  
+  struct predicate *our_pred;
+  struct exec_val *execp;	/* Pointer for efficiency. */
+
+  if ((argv == NULL) || (argv[*arg_ptr] == NULL))
+    return (false);
+
+  our_pred = insert_primary (func);
+  our_pred->side_effects = true;
+  our_pred->no_default_print = true;
+  execp = &our_pred->args.exec_vec;
+
+  if ((func != pred_okdir) && (func != pred_ok))
+    allow_plus = true;
+  else
+    allow_plus = false;
+  
+  if ((func == pred_execdir) || (func == pred_okdir))
+    {
+      check_path_safety(action);
+      execp->use_current_dir = true;
+    }
+  else
+    {
+      execp->use_current_dir = false;
+    }
+  
+  our_pred->args.exec_vec.multiple = 0;
+
+  /* Count the number of args with path replacements, up until the ';'. 
+   * Also figure out if the command is terminated by ";" or by "+".
+   */
+  start = *arg_ptr;
+  for (end = start, saw_braces=0, brace_count=0;
+       (argv[end] != NULL)
+       && ((argv[end][0] != ';') || (argv[end][1] != '\0'));
+       end++)
+    {
+      /* For -exec and -execdir, "{} +" can terminate the command. */
+      if ( allow_plus
+	   && argv[end][0] == '+' && argv[end][1] == 0
+	   && saw_braces)
+	{
+	  our_pred->args.exec_vec.multiple = 1;
+	  break;
+	}
+      
+      saw_braces = 0;
+      if (strstr (argv[end], "{}"))
+	{
+	  saw_braces = 1;
+	  ++brace_count;
+	  
+	  if (0 == end && (func == pred_execdir || func == pred_okdir))
+	    {
+	      /* The POSIX standard says that {} replacement should
+	       * occur even in the utility name.  This is insecure
+	       * since it means we will be executing a command whose
+	       * name is chosen according to whatever find finds in
+	       * the filesystem.  That can be influenced by an
+	       * attacker.  Hence for -execdir and -okdir this is not
+	       * allowed.  We can specify this as those options are 
+	       * not defined by POSIX.
+	       */
+	      error(1, 0, _("You may not use {} within the utility name for -execdir and -okdir, because this is a potential security problem."));
+	    }
+	}
+    }
+  
+  /* Fail if no command given or no semicolon found. */
+  if ((end == start) || (argv[end] == NULL))
+    {
+      *arg_ptr = end;
+      free(our_pred);
+      return false;
+    }
+
+  if (our_pred->args.exec_vec.multiple && brace_count > 1)
+    {
+	
+      const char *suffix;
+      if (func == pred_execdir)
+	suffix = "dir";
+      else
+	suffix = "";
+
+      error(1, 0,
+	    _("Only one instance of {} is supported with -exec%s ... +"),
+	    suffix);
+    }
+
+  /* execp->ctl   = xmalloc(sizeof struct buildcmd_control); */
+  bc_init_controlinfo(&execp->ctl);
+  execp->ctl.exec_callback = launch;
+
+  if (our_pred->args.exec_vec.multiple)
+    {
+      /* "+" terminator, so we can just append our arguments after the
+       * command and initial arguments.
+       */
+      execp->replace_vec = NULL;
+      execp->ctl.replace_pat = NULL;
+      execp->ctl.rplen = 0;
+      execp->ctl.lines_per_exec = 0; /* no limit */
+      execp->ctl.args_per_exec = 0; /* no limit */
+      
+      /* remember how many arguments there are */
+      execp->ctl.initial_argc = (end-start) - 1;
+
+      /* execp->state = xmalloc(sizeof struct buildcmd_state); */
+      bc_init_state(&execp->ctl, &execp->state, execp);
+  
+      /* Gather the initial arguments.  Skip the {}. */
+      for (i=start; i<end-1; ++i)
+	{
+	  bc_push_arg(&execp->ctl, &execp->state,
+		      argv[i], strlen(argv[i])+1,
+		      NULL, 0,
+		      1);
+	}
+    }
+  else
+    {
+      /* Semicolon terminator - more than one {} is supported, so we
+       * have to do brace-replacement.
+       */
+      execp->num_args = end - start;
+      
+      execp->ctl.replace_pat = "{}";
+      execp->ctl.rplen = strlen(execp->ctl.replace_pat);
+      execp->ctl.lines_per_exec = 0; /* no limit */
+      execp->ctl.args_per_exec = 0; /* no limit */
+      execp->replace_vec = xmalloc(sizeof(char*)*execp->num_args);
+
+
+      /* execp->state = xmalloc(sizeof(*(execp->state))); */
+      bc_init_state(&execp->ctl, &execp->state, execp);
+
+      /* Remember the (pre-replacement) arguments for later. */
+      for (i=0; i<execp->num_args; ++i)
+	{
+	  execp->replace_vec[i] = argv[i+start];
+	}
+    }
+  
+  if (argv[end] == NULL)
+    *arg_ptr = end;
+  else
+    *arg_ptr = end + 1;
+  
+  return true;
+}
+#else
+/* handles both exec and ok predicate */
+static boolean
+old_insert_exec_ok (boolean (*func) (/* ??? */), char **argv, int *arg_ptr)
 {
   int start, end;		/* Indexes in ARGV of start & end of cmd. */
   int num_paths;		/* Number of args with path replacements. */
@@ -1847,6 +2052,8 @@ insert_exec_ok (boolean (*func) (/* ??? */), char **argv, int *arg_ptr)
   our_pred->side_effects = true;
   our_pred->no_default_print = true;
   execp = &our_pred->args.exec_vec;
+  execp->usercontext = our_pred;
+  execp->use_current_dir = false;
   execp->paths =
     (struct path_arg *) xmalloc (sizeof (struct path_arg) * (num_paths + 1));
   execp->vec = (char **) xmalloc (sizeof (char *) * (end - start + 1));
@@ -1882,6 +2089,22 @@ insert_exec_ok (boolean (*func) (/* ??? */), char **argv, int *arg_ptr)
     *arg_ptr = end + 1;
   return (true);
 }
+#endif
+
+
+
+static boolean
+insert_exec_ok (const char *action,
+		boolean (*func) (/* ??? */), char **argv, int *arg_ptr)
+{
+#if defined(NEW_EXEC)
+  return new_insert_exec_ok(action, func, argv, arg_ptr);
+#else
+  return old_insert_exec_ok(func, argv, arg_ptr);
+#endif
+}
+
+
 
 /* Get a number of days and comparison type.
    STR is the ASCII representation.
