@@ -23,12 +23,16 @@
 #include <signal.h>
 #include <pwd.h>
 #include <grp.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 #include "../gnulib/lib/xalloc.h"
 #include "../gnulib/lib/dirname.h"
 #include "../gnulib/lib/human.h"
 #include "modetype.h"
 #include "filemode.h"
 #include "wait.h"
+#include "buildcmd.h"
 
 #if ENABLE_NLS
 # include <libintl.h>
@@ -139,7 +143,7 @@ extern int yesno ();
 #define MAX(a, b) ((a) > (b) ? (a) : (b))
 
 static boolean insert_lname PARAMS((char *pathname, struct stat *stat_buf, struct predicate *pred_ptr, boolean ignore_case));
-static boolean launch PARAMS((struct predicate *pred_ptr));
+
 static char *format_date PARAMS((time_t when, int kind));
 static char *ctime_format PARAMS((time_t when));
 
@@ -164,6 +168,7 @@ struct pred_assoc pred_table[] =
   {pred_delete, "delete  "},
   {pred_empty, "empty   "},
   {pred_exec, "exec    "},
+  {pred_execdir, "execdir "},
   {pred_false, "false   "},
   {pred_fprint, "fprint  "},
   {pred_fprint0, "fprint0 "},
@@ -186,6 +191,7 @@ struct pred_assoc pred_table[] =
   {pred_nogroup, "nogroup "},
   {pred_nouser, "nouser  "},
   {pred_ok, "ok      "},
+  {pred_okdir, "okdir   "},
   {pred_open, "(       "},
   {pred_or, "or      "},
   {pred_path, "path    "},
@@ -450,8 +456,55 @@ pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
     return (false);
 }
 
-boolean
-pred_exec (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+#if defined(NEW_EXEC)
+static boolean
+new_impl_pred_exec (const char *pathname, struct stat *stat_buf,
+		    struct predicate *pred_ptr,
+		    const char *prefix, size_t pfxlen)
+{
+  struct exec_val *execp = &pred_ptr->args.exec_vec;
+  size_t len = strlen(pathname);
+  
+  if (execp->multiple)
+    {
+      /* Push the argument onto the current list. 
+       * The command may or may not be run at this point, 
+       * depending on the command line length limits.
+       */
+      bc_push_arg(&execp->ctl,
+		  &execp->state,
+		  pathname, len+1,
+		  prefix, pfxlen,
+		  0);
+      
+      /* POSIX: If the primary expression is punctuated by a plus
+       * sign, the primary shall always evaluate as true
+       */
+      return true;
+    }
+  else
+    {
+      int i;
+
+      for (i=0; i<execp->num_args; ++i)
+	{
+	  bc_do_insert(&execp->ctl,
+		       &execp->state,
+		       execp->replace_vec[i],
+		       strlen(execp->replace_vec[i]),
+		       prefix, pfxlen,
+		       pathname, len,
+		       0);
+	}
+
+      /* Actually invoke the command. */
+      return  execp->ctl.exec_callback(&execp->ctl,
+					&execp->state);
+    }
+}
+#else
+static boolean
+old_impl_pred_exec (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
   int i;
   int path_pos;
@@ -490,6 +543,24 @@ pred_exec (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 
   return (i);
 }
+#endif
+
+boolean
+pred_exec (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+#if defined(NEW_EXEC)
+  return new_impl_pred_exec(pathname, stat_buf, pred_ptr, NULL, 0);
+#else
+  return old_impl_pred_exec(pathname, stat_buf, pred_ptr);
+#endif
+}
+
+boolean
+pred_execdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  const char *s = basename(pathname);
+  return new_impl_pred_exec(s, stat_buf, pred_ptr, "./", 2);
+}
 
 boolean
 pred_false (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
@@ -1067,6 +1138,45 @@ pred_nouser (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 #endif
 }
 
+#if defined(NEW_EXEC)
+
+static boolean
+is_ok(const char *program, const char *arg)
+{
+  fflush (stdout);
+  /* The draft open standard requires that, in the POSIX locale,
+     the last non-blank character of this prompt be '?'.
+     The exact format is not specified.
+     This standard does not have requirements for locales other than POSIX
+  */
+  fprintf (stderr, _("< %s ... %s > ? "), program, arg);
+  fflush (stderr);
+  return yesno();
+}
+
+boolean
+pred_ok (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  if (is_ok(pred_ptr->args.exec_vec.replace_vec[0], pathname))
+    return new_impl_pred_exec (pathname, stat_buf, pred_ptr, NULL, 0);
+  else
+    return false;
+}
+
+boolean
+pred_okdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  const char *s = basename(pathname);
+  
+  if (is_ok(pred_ptr->args.exec_vec.replace_vec[0], pathname))
+    return new_impl_pred_exec (s, stat_buf, pred_ptr, "./", 2);
+  else
+    return false;
+}
+
+
+
+#else
 boolean
 pred_ok (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
@@ -1085,6 +1195,17 @@ pred_ok (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
     return (false);
 }
 
+boolean
+pred_okdir (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  error(1, 0, "-okdir is not yet implemented.");
+  return false;
+}
+
+
+#endif
+
+
 boolean
 pred_open (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
@@ -1187,7 +1308,16 @@ pred_quit (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   (void) pathname;
   (void) stat_buf;
   (void) pred_ptr;
-  exit(0);
+
+  /* Run any cleanups.  This includes executing any command lines 
+   * we have partly built but not executed.
+   */
+  cleanup();
+  
+  /* Since -exec and friends don't leave child processes running in the 
+   * background, there is no need to wait for them here.
+   */
+  exit(state.exit_status);	/* 0 for success, etc. */
 }
 
 boolean
@@ -1385,6 +1515,129 @@ pred_xtype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
     zero, and the exit arg (status high) is 0.
     Otherwise return false, possibly printing an error message. */
 
+#if defined(NEW_EXEC)
+static void
+prep_child_for_exec (void)
+{
+  const char inputfile[] = "/dev/null";
+  /* fprintf(stderr, "attaching stdin to /dev/null\n"); */
+  
+  close(0);
+  if (open(inputfile, O_RDONLY) < 0)
+    {
+      /* This is not entirely fatal, since 
+       * executing the child with a closed
+       * stdin is almost as good as executing it
+       * with its stdin attached to /dev/null.
+       */
+      error (0, errno, "%s", inputfile);
+    }
+}
+
+
+
+int
+launch (const struct buildcmd_control *ctl,
+	struct buildcmd_state *buildstate)
+{
+  int wait_status;
+  pid_t child_pid;
+  static int first_time = 1;
+  const struct exec_val *execp = buildstate->usercontext;
+  
+  /* Null terminate the arg list.  */
+  bc_push_arg (ctl, buildstate, (char *) NULL, 0, NULL, 0, false); 
+  
+  /* Make sure output of command doesn't get mixed with find output. */
+  fflush (stdout);
+  fflush (stderr);
+  
+  /* Make sure to listen for the kids.  */
+  if (first_time)
+    {
+      first_time = 0;
+      signal (SIGCHLD, SIG_DFL);
+    }
+
+  child_pid = fork ();
+  if (child_pid == -1)
+    error (1, errno, _("cannot fork"));
+  if (child_pid == 0)
+    {
+      /* We be the child. */
+      prep_child_for_exec();
+
+      /* For -exec and -ok, change directory back to the starting directory.
+       * for -execdir and -okdir, stay in the directory we are searching
+       * (the latter is more secure).
+       */
+      if (!execp->use_current_dir)
+	{
+	  if (starting_desc < 0
+	      ? chdir (starting_dir) != 0
+	      : fchdir (starting_desc) != 0)
+	    {
+	      error (0, errno, "%s", starting_dir);
+	      _exit (1);
+	    }
+	}
+      
+      execvp (buildstate->cmd_argv[0], buildstate->cmd_argv);
+      error (0, errno, "%s", buildstate->cmd_argv[0]);
+      _exit (1);
+    }
+
+
+  /* In parent; set up for next time. */
+  bc_clear_args(ctl, buildstate);
+
+  
+  while (waitpid (child_pid, &wait_status, 0) == (pid_t) -1)
+    {
+      if (errno != EINTR)
+	{
+	  error (0, errno, _("error waiting for %s"), buildstate->cmd_argv[0]);
+	  state.exit_status = 1;
+	  return 0;		/* FAIL */
+	}
+    }
+  
+  if (WIFSIGNALED (wait_status))
+    {
+      error (0, 0, _("%s terminated by signal %d"),
+	     buildstate->cmd_argv[0], WTERMSIG (wait_status));
+      
+      if (execp->multiple)
+	{
+	  /* -exec   \; just returns false if the invoked command fails. 
+	   * -exec {} + returns true if the invoked command fails, but
+	   *            sets the program exit status.
+	   */
+	  state.exit_status = 1;
+	}
+      
+      return 1;			/* OK */
+    }
+
+  if (0 == WEXITSTATUS (wait_status))
+    {
+      return 1;			/* OK */
+    }
+  else
+    {
+      if (execp->multiple)
+	{
+	  /* -exec   \; just returns false if the invoked command fails. 
+	   * -exec {} + returns true if the invoked command fails, but
+	   *            sets the program exit status.
+	   */
+	  state.exit_status = 1;
+	}
+      return 0;			/* FAIL */
+    }
+  
+}
+#else
 static boolean
 launch (struct predicate *pred_ptr)
 {
@@ -1441,6 +1694,8 @@ launch (struct predicate *pred_ptr)
     }
   return (!WEXITSTATUS (status));
 }
+#endif
+
 
 /* Return a static string formatting the time WHEN according to the
    strftime format character KIND.  */
